use std::{
    cmp::Ordering,
    ffi::{OsStr, OsString},
    fs,
    path::{Path, PathBuf},
};

use chrono::{DateTime, NaiveDateTime, Utc};
use sqlx::database;
use strum_macros::{EnumDiscriminants, EnumIter};

use crate::{database::Database, sheet_dao};

pub trait PdfSheet {
    fn get_pdf(&self) -> &Pdf;
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Sheet {
    pub id: i64,
    pub last_opened: I64DateTime,
    pub kind: SheetKind,
}

#[derive(Debug, Clone, PartialEq, Eq, EnumDiscriminants)]
#[strum_discriminants(derive(EnumIter))]
pub enum SheetKind {
    Sheet {
        pdf: Pdf,
        name: String,
        composer_id: i64,
    },
    Orphan {
        pdf: Pdf,
    },
    Book {
        pdf: Pdf,
        name: String,
        composer_id: i64,
        sheet_ids: Vec<i64>,
    },
    BookSheet {
        name: String,
        book_id: i64,
        first_page: i64,
        last_page: i64,
    },
}

impl PartialOrd for Sheet {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl Ord for Sheet {
    fn cmp(&self, other: &Self) -> Ordering {
        self.last_opened.cmp(&other.last_opened)
    }
}

impl Sheet {
    pub async fn open_file(&self, database: &Database) {
        let path = match &self.kind {
            SheetKind::Sheet { pdf, .. } => pdf.path.clone(),
            SheetKind::Orphan { pdf } => pdf.path.clone(),
            SheetKind::Book { pdf, .. } => pdf.path.clone(),
            SheetKind::BookSheet { book_id, .. } => sheet_dao::find_path_of_book(database, book_id)
                .await
                .unwrap(),
        };
        opener::open(path).unwrap();
    }

    pub fn update_pdf_file(&mut self, new_pdf: Pdf) -> std::io::Result<()> {
        match &mut self.kind {
            SheetKind::Sheet { pdf, .. } => *pdf = new_pdf,
            SheetKind::Orphan { pdf, .. } => *pdf = new_pdf,
            SheetKind::Book { pdf, .. } => *pdf = new_pdf,
            SheetKind::BookSheet { .. } => {} // TODO: find better solution!
        };
        Ok(())
    }

    pub fn pdf_path_equal(&self, path: impl AsRef<Path>) -> bool {
        match &self.kind {
            SheetKind::Sheet { pdf, .. } => pdf.path == path.as_ref(),
            SheetKind::Orphan { pdf, .. } => pdf.path == path.as_ref(),
            SheetKind::Book { pdf, .. } => pdf.path == path.as_ref(),
            SheetKind::BookSheet { .. } => false, // TODO: find better solution!
        }
    }

    pub fn validate_own_path(&self) -> std::io::Result<bool> {
        Ok(match &self.kind {
            SheetKind::Sheet { pdf, .. } => pdf.validate_path(&pdf.path)?,
            SheetKind::Orphan { pdf, .. } => pdf.validate_path(&pdf.path)?,
            SheetKind::Book { pdf, .. } => pdf.validate_path(&pdf.path)?,
            SheetKind::BookSheet { book_id, .. } => true, // TODO: better solution?
        })
    }

    pub fn try_get_path(&self) -> Option<&Path> {
        match &self.kind {
            SheetKind::Sheet { pdf, .. } => Some(&pdf.path),
            SheetKind::Orphan { pdf, .. } => Some(&pdf.path),
            SheetKind::Book { pdf, .. } => Some(&pdf.path),
            SheetKind::BookSheet { .. } => None,
        }
    }

    pub fn validate_path(&self, path: impl AsRef<Path>) -> std::io::Result<bool> {
        Ok(match &self.kind {
            SheetKind::Sheet { pdf, .. } => pdf.validate_path(path)?,
            SheetKind::Orphan { pdf, .. } => pdf.validate_path(path)?,
            SheetKind::Book { pdf, .. } => pdf.validate_path(path)?,
            SheetKind::BookSheet { book_id, .. } => true, // TODO: better solution?
        })
    }
}

impl SheetKindDiscriminants {
    pub fn get_database_table_name(&self) -> &str {
        match self {
            SheetKindDiscriminants::Sheet => "sheets",
            SheetKindDiscriminants::Orphan => "orphans",
            SheetKindDiscriminants::Book => "books",
            SheetKindDiscriminants::BookSheet => "booksheets",
        }
    }
}

#[derive(sqlx::FromRow)]
pub struct Composer {
    id: i64,
    name: String,
}

#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord)]
pub struct I64DateTime(pub DateTime<Utc>);

impl TryFrom<i64> for I64DateTime {
    type Error = String;

    fn try_from(value: i64) -> Result<Self, Self::Error> {
        Ok(I64DateTime(
            NaiveDateTime::from_timestamp_opt(value, 0)
                .ok_or("Failed converting i64 to DateTime")?
                .and_utc(),
        ))
    }
}

impl From<&I64DateTime> for i64 {
    fn from(value: &I64DateTime) -> Self {
        value.0.timestamp()
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Pdf {
    pub path: PathBuf,
    pub file_size: u64,
    pub file_hash: String,
}

impl Pdf {
    pub fn get_name(&self) -> &str {
        self.path.file_name().unwrap().to_str().unwrap()
    }

    pub fn validate_path(&self, path: impl AsRef<Path>) -> std::io::Result<bool> {
        // First compare file size since it is faster than hashing
        let file_size = fs::metadata(path.as_ref())?.len();
        if file_size == self.file_size {
            let file_content = fs::read(path.as_ref())?;
            let file_hash = blake3::hash(&file_content);
            if file_hash.to_string() == self.file_hash {
                return Ok(true);
            }
        }

        Ok(false)
    }
}

impl TryFrom<PathBuf> for Pdf {
    type Error = std::io::Error;

    fn try_from(path: PathBuf) -> Result<Self, Self::Error> {
        let file_size = fs::metadata(path.as_path())?.len();
        let file_content = fs::read(path.as_path())?;
        let file_hash = blake3::hash(&file_content).to_string();

        Ok(Pdf {
            path,
            file_size,
            file_hash,
        })
    }
}
